import 'dart:async';
import 'package:meta/meta.dart';
import 'package:rxdart/rxdart.dart';
import 'action.dart';
import 'actions.dart';
import 'types.dart';

typedef EffectCallback = Stream<Action> Function(
    Actions action$, AjwahStore store);

class AjwahStore {
  BehaviorSubject<Action> _dispatcher;
  BehaviorSubject<Map<String, dynamic>> _store;
  Map<String, StreamSubscription<Action>> _stateSubscriptions;
  Map<String, StreamSubscription<Action>> _effectSubscriptions;
  Actions _actions;
  AjwahStore() {
    _dispatcher = BehaviorSubject<Action>.seeded(Action(type: '@@INIT'));
    _store = BehaviorSubject<Map<String, dynamic>>.seeded({});
    _actions = Actions(_dispatcher);
    _stateSubscriptions = <String, StreamSubscription<Action>>{};
    _effectSubscriptions = <String, StreamSubscription<Action>>{};
  }
  void registerState<S>({
    @required String stateName,
    @required S initialState,
    FilterActionCallback filterActions,
    @required MapActionToStateCallback<S> mapActionToState,
  }) {
    if (_store.value.containsKey(stateName)) {
      return;
    }
    _store.value[stateName] = initialState;
    _store.add(_store.value);
    dispatch(Action(type: 'registerState($stateName)'));
    void emitState(S state) {
      if (_store.value[stateName] != state) {
        _store.value[stateName] = state;
        _store.add(_store.value);
      }
    }

    if (filterActions is FilterActionCallback) {
      _stateSubscriptions[stateName] =
          _dispatcher.where(filterActions).listen((action) {
        mapActionToState(_store.value[stateName], action, emitState);
      });
    } else {
      _stateSubscriptions[stateName] = _dispatcher.listen((action) {
        mapActionToState(_store.value[stateName], action, emitState);
      });
    }
  }

  Stream<T> select<T>(String stateName) {
    return _store.map<T>((dic) => dic[stateName]).distinct();
  }

  Stream<T> selectMany<T>(T Function(Map<String, dynamic> state) callback) {
    return _store.map<T>(callback).distinct();
  }

  void dispatch(Action action) {
    _dispatcher.add(action);
  }

  ///This method is usefull to add a single effect passing a callback **(
  ///Actions action$, Store store$)=>Stream** and **effectKey**.
  ///
  ///**Example**
  ///```dart
  ///registerEffect('effect-key',[(action$, store$)=>action$
  ///           .whereType(ActionTypes.AsyncInc)
  ///           .debounceTime(Duration(milliseconds: 1000))
  ///           .mapTo(Action(type: ActionTypes.Inc))
  ///           ]);
  ///```
  void registerEffects(
      String effectKey, Iterable<EffectCallback> callbackList) {
    if (_effectSubscriptions.containsKey(effectKey)) {
      return;
    }
    _effectSubscriptions[effectKey] =
        Rx.merge(callbackList.map((e) => e(_actions, this)))
            .asBroadcastStream()
            .listen(dispatch);
    dispatch(Action(type: 'registerEffects($effectKey)'));
  }

  ///This method is usefull to remove effects passing **effectKey** on demand.
  void unregisterEffects({@required String effectKey}) {
    if (_effectSubscriptions.containsKey(effectKey)) {
      _effectSubscriptions[effectKey].cancel();
      _effectSubscriptions.remove(effectKey);
      dispatch(Action(type: 'unregisterEffects($effectKey)'));
    }
  }

  void unregisterState({@required String stateName}) {
    if (_stateSubscriptions.containsKey(stateName)) {
      _stateSubscriptions[stateName].cancel();
      _stateSubscriptions.remove(stateName);
      _store.value.remove(stateName);
      _store.add(_store.value);
      dispatch(Action(type: 'unregisterState($stateName)'));
    } else if (_store.value.containsKey(stateName)) {
      _store.value.remove(stateName);
      _store.add(_store.value);
      dispatch(Action(type: 'unregisterState($stateName)'));
    }
  }

  Actions get actions => _actions;
  T getState<T>({@required String stateName}) => _store.value[stateName];

  ///return latest stream of [action, state] array.
  ///
  /// **Example**
  /// ```dart
  ///store.exportState().listen((arr) {
  ///    print((arr[0] as Action).type);
  ///    print(arr[1]);
  ///  });
  /// ```
  Stream<List<dynamic>> exportState() =>
      _dispatcher.withLatestFrom(_store, (t, s) => [t, s]);

  ///state object should be a **Map<String, dynamic> state**
  ///
  /// **Example**
  /// ```dart
  /// var state={'counter':CounterModel(count:5, isLoading:false)};
  /// store.importState(state);
  /// ```
  void importState(Map<String, dynamic> state) {
    _store.add(state);
    dispatch(Action(type: '@importState'));
  }

  ///It's a clean up function.
  void dispose() {
    _stateSubscriptions.forEach((key, value) {
      value.cancel();
    });
    _effectSubscriptions.forEach((key, value) {
      value.cancel();
    });
    _stateSubscriptions.clear();
    _effectSubscriptions.clear();

    _dispatcher.close();
    _store.close();
  }
}
